# Copyright WillianFuks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Summarizes performance information inferred in post-inferences compilation process.
"""


import os

import pandas as pd
from jinja2 import Template

from causalimpact.misc import get_z_score

_here = os.path.dirname(os.path.abspath(__file__))
summary_tmpl_path = os.path.join(_here, 'summary', 'templates', 'summary')
report_tmpl_path = os.path.join(_here, 'summary', 'templates', 'report')

SUMMARY_TMPL = Template(open(summary_tmpl_path).read())
REPORT_TMPL = Template(open(report_tmpl_path).read())


def summary(
    summary_data: pd.DataFrame,
    p_value: float,
    alpha: float = 0.05,
    output: str = 'summary',
    digits: int = 2
) -> str:
    """
    Returns final results from causal impact analysis, such as absolute observed effect,
    the relative effect between prediction and observed variable, cumulative performances
    in post-intervention period among other metrics.

    Args
    ----
      summary_data: pd.DataFrame
          Contains information such as means and cumulatives averages.
      p_value: float
          p-value test for testing presence of signal in data.
      alpha: float
          Sets credible interval width.
      output: str
          Can be either "summary" or "report". The first is a simpler output just
          informing general metrics such as expected absolute or relative effect.
      digits: int
          Defines the number of digits after the decimal point to round. For `digits=2`,
          value 1.566 becomes 1.57.

    Returns
    -------
      summary: str
          Contains results of the causal impact analysis.

    Raises
    ------
      ValueError: If input `output` is not either 'summary' or 'report'.
    """
    if output not in {'summary', 'report'}:
        raise ValueError('Please choose either summary or report for output.')
    if output == 'summary':
        summary = SUMMARY_TMPL.render(
            summary=summary_data.to_dict(),
            alpha=alpha,
            z_score=get_z_score(1 - alpha / 2.),
            p_value=p_value,
            digits=digits
        )
    else:
        summary = REPORT_TMPL.render(
            summary=summary_data.to_dict(),
            alpha=alpha,
            p_value=p_value,
            digits=digits
        )
    return summary



# Copyright WillianFuks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module responsible for all functions related to processing model certification and
creation.
"""


from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import tensorflow as tf
import tensorflow_probability as tfp

tfd = tfp.distributions
tfb = tfp.bijectors

# K Local Level Prior Sample Size
# This is equal to the original [R package](https://github.com/google/CausalImpact/blob/07b60e1bf5c9c8d74e31ea602db39d7256a53b6f/R/impact_model.R#L25) # noqa: E501
kLocalLevelPriorSampleSize = 32


def process_model_args(model_args: Dict[str, Any]) -> Dict[str, Any]:
    """
    Process general parameters related to how Causal Impact will be implemented, such
    as standardization procedure or the addition of seasonal components to the model.

    Args
    ----
      model_args:
        niter: int
            How many iterations to run either for `hmc` or `vi` algorithms.
        standardize: bool
            If `True`, standardize data so result has zero mean and unitary standard
            deviation.
        prior_level_sd: float
            Standard deviation that sets initial local level distribution. Default
            value is 0.01 which means the linear regression is expected to explain
            well the observed data. In cases where this is not expected, then it's also
            possible to use the value 0.1. Still, this value will increase considerably
            the extension of the random walk variance modeling data which can lead to
            unreliable predictions (this might indicate that better covariates are
            required).
        fit_method: str
            Which method to use when fitting the structural time series model. Can be
            either `hmc` which stands for "Hamiltonian Monte Carlo" or "vi", i.e.,
            "variational inference". The first is slower but more accurate whilst the
            latter is the opposite. Defaults to `vi` which prioritizes performance.
        nseasons: int
            Specifies the duration of the period of the seasonal component; if input
            data is specified in terms of days, then choosing nseasons=7 adds a weekly
            seasonal effect.
        season_duration: int
            Specifies how many data points each value in season spans over. A good
            example to understand this argument is to consider a hourly data as input.
            For modeling a weekly season on this data, one can specify `nseasons=7` and
            season_duration=24 which means each value that builds the season component
            is repeated for 24 data points. Default value is 1 which means the season
            component spans over just 1 point (this in practice doesn't change
            anything). If this value is specified and bigger than 1 then `nseasons`
            must be specified and bigger than 1 as well.

    Returns
    -------
        Dict[str, Any]
          standardize: bool
          prior_level_sd: float
          niter: int
          fit_method: str
          nseasons: int
          season_duration: int

    Raises
    ------
      ValueError: if `standardize` is not of type `bool`.
                  if `prior_level_sd` is not `float`.
                  if `niter` is not `int`.
                  if `fit_method` not in {'hmc', 'vi'}.
                  if `nseasons` is not `int`.
                  if `season_duration` is not `int`.
                  if `season_duration` is bigger than 1 and `nseasons` is 1.
    """
    standardize = model_args.get('standardize', True)
    if not isinstance(standardize, bool):
        raise ValueError('standardize argument must be of type bool.')
    model_args['standardize'] = standardize

    prior_level_sd = model_args.get('prior_level_sd', 0.01)
    if not isinstance(prior_level_sd, float):
        raise ValueError('prior_level_sd argument must be of type float.')
    model_args['prior_level_sd'] = prior_level_sd

    niter = model_args.get('niter', 1000)
    if not isinstance(niter, int):
        raise ValueError('niter argument must be of type int.')
    model_args['niter'] = niter

    fit_method = model_args.get('fit_method', 'vi')
    if fit_method not in {'hmc', 'vi'}:
        raise ValueError('fit_method can be either "hmc" or "vi".')
    model_args['fit_method'] = fit_method

    nseasons = model_args.get('nseasons', 1)
    if not isinstance(nseasons, int):
        raise ValueError('nseasons argument must be of type int.')
    model_args['nseasons'] = nseasons

    season_duration = model_args.get('season_duration', 1)
    if not isinstance(season_duration, int):
        raise ValueError('season_duration argument must be of type int.')
    if nseasons <= 1 and season_duration > 1:
        raise ValueError('nseasons must be bigger than 1 when season_duration is also '
                         'bigger than 1.')
    model_args['season_duration'] = season_duration
    return model_args


def check_input_model(
    model: tfp.sts.StructuralTimeSeries,
    pre_data: pd.DataFrame,
    post_data: pd.DataFrame
) -> None:
    """
    Checkes whether input model was properly built and is ready to be run. This function
    is only invoked if the client sent a customized input model. Various assertions are
    performed to guarantee it has been created appropriately, such as each component
    should have `len(pre_data)` points for the argument `observed_time_series`. In case
    the component is of type `tfp.sts.LinearRegression` or `SparseLinearRegression` then
    the design matrix must have
    `shape = (len(pre_data) + len(post_data), cols(pre_data) - 1)` which allows not only
    to fit the model as well as to run the forecasts.
    The model must be built with data of dtype=tf.float32 or np.float32 as otherwise an
    error will be thrown when fitting the markov chains.

    Args
    ----
      model: StructuralTimeSeries
          Can be either default `LocalLevel` or user specified generic model.
      pre_data: pd.DataFrame

    Raises
    ------
      ValueError: if model is not of appropriate type.
                  if model is built without appropriate observed time series data.
                  if model components don't have dtype=tf.float32 or np.float32
    """
    def _check_component(component):
        if isinstance(
            component,
            (tfp.sts.LinearRegression, tfp.sts.SparseLinearRegression)
        ):
            covariates_shape = (len(pre_data) + len(post_data),
                                len(pre_data.columns) - 1)
            if component.design_matrix.shape != covariates_shape:
                raise ValueError(
                    'Customized Linear Regression Models must have total '
                    'points equal to pre_data and post_data points and '
                    'same number of covariates. Input design_matrix shape was '
                    f'{component.design_matrix.shape} and expected '
                    f'{(len(pre_data) + len(post_data), len(pre_data.columns) -1)} '
                    'instead.'
                )
            assert component.design_matrix.dtype == tf.float32
        else:
            for parameter in component.parameters:
                assert parameter.prior.dtype == tf.float32

    if not isinstance(model, tfp.sts.StructuralTimeSeries):
        raise ValueError('Input model must be of type StructuralTimeSeries.')
    if isinstance(model, tfp.sts.Sum):
        for component in model.components:
            _check_component(component)
    else:
        _check_component(model)


def build_inv_gamma_sd_prior(sigma_guess: float) -> tfd.Distribution:
    """
    helper function to build the sd_prior distribution for standard deviation
    modeling.

    Args
    ----
      sigma_guess: float
          Initial guess of where the standard deviation of the parameter is located.

    Returns
    -------
      tfd.Distribution
          InverseGamma distribution modeling the standard deviation.
    """
    sample_size = kLocalLevelPriorSampleSize
    df = sample_size
    a = np.float32(df / 2)
    ss = sample_size * sigma_guess ** 2
    b = np.float32(ss / 2)
    return tfd.InverseGamma(a, b)


def build_bijector(dist: tfd.Distribution) -> tfd.Distribution:
    """
    helper function for building final bijector given sd_prior. The bijector is
    implemented through the `tfd.TransformedDistribution` class.

    Args
    ----
      dist: tfd.Distribution
          Distribution to receive the transformation `G(X)`.

    Returns
    -------
      new_dist: tfd.Distribution
          New distribution given by `y = G(X)`.
    """
    bijector = SquareRootBijector()
    new_dist = tfd.TransformedDistribution(dist, bijector)
    return new_dist


def build_default_model(
    observed_time_series: pd.DataFrame,
    pre_data: pd.DataFrame,
    post_data: pd.DataFrame,
    prior_level_sd: float,
    nseasons: int,
    season_duration: int
) -> tfp.sts.StructuralTimeSeries:
    """
    When input model is `None` then proceeds to build a default `tfp.sts.LocalLevel`
    model. If input data has covariates then also adds a `tfp.sts.SparseLinearRegression`
    component.

    The level_prior follows `1 / prior_level_sd ** 2 ~ Gamma(a, b)` according to
    the original [BOOM](https://github.com/steve-the-bayesian/BOOM/blob/63f08a708153c8405b809405fa1ab5ed7193d648/Interfaces/python/R/R/bayes.py#L4:L12) package.  # noqa: E501
    This is achieved by using the InverseGamma(a, b) and a [bijector](https://tiao.io/post/building-probability-distributions-with-tensorflow-probability-bijector-api/) # noqa: E501
    transformation for the square root operator.

    As for the linear regressor, the `tfp.sts.SparseLinearRegression` operation is similar
    to the spike-and-slab from the original R package; main difference is that it follows
    instead a horseshoe distribution which tends to penalize less the meaningful weights
    in the shrinking process.[https://github.com/tensorflow/probability/blob/v0.12.1/tensorflow_probability/python/sts/regression.py#L265-L523] # noaq: E501

    Args
    ----
      observed_time_series: pd.DataFrame
      pre_data: pd.DataFrame
      post_data: pd.DataFrame
      prior_level_sd: float
          Sets an initial estimation for the standard deviation 'sigma' of the local
          level prior. The bigger this value is, the wider is expected to be the random
          walk extension on forecasts.
      nseasons: int
      season_duration: int

    Returns
    -------
      model: tfp.sts.Sum
          A `tfp.sts.LocalLevel` default model with possible another
          `tfp.sts.SparseLinearRegression` and `tfp.sts.Seasonal` components representing
          covariates and seasonal patterns.
    """
    components = []
    obs_sd = observed_time_series.values.std()  # use `values` to avoid batching dims
    sd_prior = build_inv_gamma_sd_prior(prior_level_sd)
    sd_prior = build_bijector(sd_prior)
    # This is an approximation to simulate the bsts package from R. It's expected that
    # given a few data points the posterior will converge appropriately given this
    # distribution, that's why it's divided by 2.
    obs_prior = build_inv_gamma_sd_prior(obs_sd / 2)
    obs_prior = build_bijector(obs_prior)
    level_component = tfp.sts.LocalLevel(
        level_scale_prior=sd_prior,
        observed_time_series=observed_time_series
    )
    components.append(level_component)
    # If it has more than 1 column then it has covariates X so add a linear regressor
    # component.
    if len(pre_data.columns) > 1:
        # We need to concatenate both pre and post data as this will allow the linear
        # regressor component to use the post data when running forecasts. As first
        # column is supposed to have response variable `y` then we filter out just the
        # remaining columns for the `X` value.
        complete_data = pd.concat([pre_data, post_data]).astype(np.float32)
        linear_component = tfp.sts.SparseLinearRegression(
            design_matrix=complete_data.iloc[:, 1:]
        )
        components.append(linear_component)
    if nseasons > 1:
        seasonal_component = tfp.sts.Seasonal(
            num_seasons=nseasons,
            num_steps_per_season=season_duration,
            observed_time_series=observed_time_series
        )
        components.append(seasonal_component)
    # Model must be built with `tfp.sts.Sum` so to add the observed noise `epsilon`
    # parameter.
    model = tfp.sts.Sum(components, observed_time_series=observed_time_series,
                        observation_noise_scale_prior=obs_prior)
    return model


def fit_model(
    model: tfp.sts.StructuralTimeSeries,
    observed_time_series: pd.DataFrame,
    method: str = 'hmc'
) -> Tuple[Union[List[tf.Tensor], Dict[str, tf.Tensor]], Optional[Dict[str, Any]]]:
    """
    Run the Markovian Monte Carlo fitting process for finding the posterior `P(z | y)`
    where z represents the structural components of the input state space model. Two
    main methods can be used, either `hmc` which stands for 'Hamiltonian Monte Carlo'
    and `vi` standing for 'Variational Inference'. The first method is expected to be
    more accurate while less performante whereas the second is the opposite, that is,
    faster but less accurate.

    Args
    ----
      model: tfp.sts.StructuralTimeSeries
          Structural time series model built to explain the observed data. It may
          contain several components such as local level, seasons and so on.
      observed_time_series: pd.DataFrame
          Contains the pre-period response variable `y`.
      method: str
          Either 'hmc' or 'vi' which selects which fitting process to run.

    Returns
    -------
      (samples, kernel_results): Tuple[Union[List[tf.Tensor], Dict[str, tf.Tensor]],
                                       Dict[str, Any]]

    Raises
    ------
      ValueError: If input method is invalid.
    """
    if method == 'hmc':
        # this method does not need to be wrapped in a `tf.function` context as the
        # internal sampling method already is:
        # https://github.com/tensorflow/probability/blob/v0.11.1/tensorflow_probability/python/sts/fitting.py#L422 # noqa: E501
        # https://github.com/tensorflow/probability/issues/348
        samples, kernel_results = tfp.sts.fit_with_hmc(
            model=model,
            observed_time_series=observed_time_series,
        )
        return samples, kernel_results
    elif method == 'vi':
        optimizer = tf.optimizers.Adam(learning_rate=0.1)
        variational_steps = 200  # Hardcoded for now
        variational_posteriors = tfp.sts.build_factored_surrogate_posterior(model=model)

        @tf.function()
        def _run_vi():  # pragma: no cover
            tfp.vi.fit_surrogate_posterior(
                target_log_prob_fn=model.joint_log_prob(
                    observed_time_series=observed_time_series
                ),
                surrogate_posterior=variational_posteriors,
                optimizer=optimizer,
                num_steps=variational_steps
            )
            # Don't sample too much as varitional inference method is built aiming for
            # performance first.
            samples = variational_posteriors.sample(100)
            return samples, None
        return _run_vi()
    else:
        raise ValueError(
            f'Input method "{method}" not valid. Choose between "hmc" or "vi".'
        )


def build_one_step_dist(
    model: tfp.sts.StructuralTimeSeries,
    observed_time_series: pd.DataFrame,
    parameter_samples: Union[List[tfd.Distribution], Dict[str, tfd.Distribution]]
) -> tfd.Distribution:  # pragma: no cover
    """
    Builds one step distribution for pre-intervention data given samples from the
    posterior `P(z | y)`.

    Args
    ----
      model: tfp.StructuralTimeSeries
      observed_time_series: pd.DataFrame
          Corresponds to the `y` value.
      parameter_samples: Union[List[tfd.Distribution], Dict[str, tfd.Distribution]]
          samples from the posterior for each state component in `model`.

    Returns
    -------
      one_step_dist: tfd.Distribution
    """
    return tfp.sts.one_step_predictive(
        model=model,
        observed_time_series=observed_time_series,
        parameter_samples=parameter_samples
    )


def build_posterior_dist(
    model: tfp.sts.StructuralTimeSeries,
    observed_time_series: pd.DataFrame,
    parameter_samples: Union[List[tfd.Distribution], Dict[str, tfd.Distribution]],
    num_steps_forecast: int
) -> tfd.Distribution:  # pragma: no cover
    """
    Builds the distribution for post-intervention data given samples from the
    posterior `P(z | y)`.

    Args
    ----
      model: tfp.StructuralTimeSeries
      observed_time_series: pd.DataFrame
          Corresponds to the `y` value.
      parameter_samples: Union[List[tfd.Distribution], Dict[str, tfd.Distribution]]
          samples from the posterior for each state component in `model`.
      num_steps_forecast: int
          How many time steps to forecast into the future. These will be compared against
          the real value of `y` to extract the estimation of impact.

    Returns
    -------
      posterior_dist: tfd.Distribution
    """
    return tfp.sts.forecast(
        model=model,
        observed_time_series=observed_time_series,
        parameter_samples=parameter_samples,
        num_steps_forecast=num_steps_forecast
    )


class SquareRootBijector(tfb.Bijector):
    """
    Compute `Y = g(X) = X ** (1 / 2) which transforms variance into standard deviation.
    Main reference for building this bijector is the original [PowerTransform](https://github.com/tensorflow/probability/blob/v0.11.1/tensorflow_probability/python/bijectors/power_transform.py) # noqa: E501
    """
    def __init__(
        self,
        validate_args: bool = False,
        name: str = 'square_root_bijector'
    ):
        """
        Args
        ----
          validate_args: bool
              Indicates whether arguments should be checked for correctness.
          name: str
              Name given to ops managed by this object.
        """
        # Without these `parameters` the code won't be compatible with future versions
        # of tfp:
        # https://github.com/tensorflow/probability/issues/1202
        parameters = dict(locals())
        with tf.name_scope(name) as name:
            super().__init__(
                forward_min_event_ndims=0,
                validate_args=validate_args,
                parameters=parameters,
                name=name)

    def _forward(self, x: Union[float, np.array, tf.Tensor]) -> tf.Tensor:
        """
        Implements the forward pass `G` as given by `Y = G(X)`. In this case, it's a
        simple square root of X.

        Args
        ----
          x: Union[float, np.array, tf.Tensor])
              Variable `X` to receive the transformation.

        Returns
        -------
          X: tf.Tensor
              Square root of `x`.
        """
        return tf.sqrt(x)

    def _inverse(self, y: Union[float, np.array, tf.Tensor]) -> tf.Tensor:
        """
        Implements G^-1(y).

        Args
        ----
          y: Union[float, np.array, tf.Tensor]
              Values to be transformed back. In this case, they will be squared.

        Returns
        -------
          y: tf.Tensor
              Squared `y`.
        """
        return tf.square(y)

    def _inverse_log_det_jacobian(self, y: tf.Tensor) -> tf.Tensor:
        """
        When transforming from `P(X)` to `P(Y)` it's necessary to compute the log of the
        determinant of the Jacobian matrix for each correspondent function `G` which
        accounts for the volumetric transformations on each domain.

        The inverse log determinant is given by:

        `ln(|J(G^-1(Y)|) = ln(|J(Y ** 2)|) = ln(|2 * Y|) = ln(2 * Y)`

        Args
        ----
          y: tf.Tensor

        Returns
        -------
          tf.Tensor
        """
        return tf.math.log(2 * y)

    def _forward_log_det_jacobian(self, x: tf.Tensor) -> tf.Tensor:
        """
        Computes the volumetric change when moving forward from `P(X)` to `P(Y)`, given
        by:

        `ln(|J(G(X))|) = ln(|J(sqrt(X))|) = ln(|(1 / 2) * X ** (-1 / 2)|) =
                       = (-1 / 2) * ln(4.0 * X)

        Args
        ----
          x: tf.Tensor

        Returns
        -------
          tf.tensor
        """
        return -0.5 * tf.math.log(4.0 * x)



# Copyright WillianFuks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Main class definition for running Causal Impact analysis.
"""

from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import tensorflow_probability as tfp

import causalimpact.data as cidata
import causalimpact.inferences as inferrer
import causalimpact.model as cimodel
import causalimpact.plot as plotter
import causalimpact.summary as summarizer
from causalimpact.misc import maybe_unstandardize


class CausalImpact():
    """
    Main class used to run the Causal Impact algorithm implemented by Google as
    described in the offical
    [paper](https://google.github.io/CausalImpact/CausalImpact.html).

    The algorithm basically fits a structural state space model to observed data `y` and
    uses Bayesian inferencing to find the posterior P(z|y) where `z` represents for the
    chosen model parameters (such as level, trend, season, and so on).

    In this package, the fitting method can be either 'Hamitonian Monte Carlo' or 'hmc'
    for short (more accurate algorithm but slower) or 'Variational Inference' or 'vi'
    (faster but less accurate), both available on Tensorflow Probability.

    Args
    ----
      data: Union[np.array, pd.DataFrame]
          First column must contain the `y` value whose future values will be forecasted
          while the remaining data contains the covariates `X` that are used in the
          linear regression component of the model (supposing that there's a linear
          regression otherwise `X` is not specified).
          If `data` it's a pandas DataFrame, its index can be defined either as a
          `RangeIndex`, `Index` or `DateTimeIndex`.
          In case of the second, then a conversion to `DateTime` type is automatically
          performed; in case of failure, the original index is kept untouched.
      pre_period: Union[List[int], List[str], List[pd.Timestamp]]
          A list of size two containing either `int`, `str` or `pd.Timestamp` values
          that references the range from beginning to end to be used in the
          pre-intervention data.
          As an example, valid inputs are:
            - [0, 30]
            - ['20200101', '20200130']
            - [pd.to_datetime('20200101'), pd.to_datetime('20200130')]
            - [pd.Timestamp('20200101'), pd.Timestamp('20200130')]
          The latter can be used only if the input `data` is a pandas DataFrame whose
          index is based on datetime values.
      post_period: Union[List[int], List[str], List[pd.Timestamp]]
          The same as `pre_period` but references where the post-intervention
          data begins and ends. This is the data that will be compared against the
          counter-factual forecasts.
      model: Optional[tfp.sts.StructuralTimeSeries]
          If `None` then a default `tfp.sts.LocalLevel` model is internally built
          otherwise use the input `model` for fitting and forecasting.
      model_args: Dict[str, Any]
          Sets general variables for building and running the state space model. Possible
          values are:
            standardize: bool
                If `True`, standardizes data to have zero mean and unitary standard
                deviation.
            prior_level_sd: Optional[float]
                Prior value for the local level standard deviation. If `None` then an
                automatic optimization of the local level is performed. This is
                recommended when there's uncertainty about what prior value is
                appropriate for the data.
                In general, if the covariates are expected to be good descriptors of the
                observed response then this value can be low (such as the default of
                0.01). In cases when the linear regression is not quite expected to fully
                explain the observed data, the value 0.1 can be used.
            fit_method: str
                Which method to use for the Bayesian algorithm. Can be either 'vi'
                (default) or 'hmc' (more precision but much slower).
            nseasons: int
              Specifies the duration of the period of the seasonal component; if input
              data is specified in terms of days, then choosing nseasons=7 adds a weekly
              seasonal effect.
            season_duration: int
              Specifies how many data points each value in season spans over. A good
              example to understand this argument is to consider a hourly data as input.
              For modeling a weekly season on this data, one can specify `nseasons=7` and
              season_duration=24 which means each value that builds the season component
              is repeated for 24 data points. Default value is 1 which means the season
              component spans over just 1 point (this in practice doesn't change
              anything). If this value is specified and bigger than 1 then `nseasons`
              must be specified and bigger than 1 as well.
      alpha: float
          A float that ranges between 0 and 1 indicating the significance level that
          will be used when statistically testing for signal presencen in the post-
          intervention period.

    Returns
    -------
      Causal Impact object with inferences, summary and plotting functionalities.

    Examples
    --------

      Imput data can be a `numpy.array`:

      ```python
      import numpy as np


      data = np.random.rand(100, 2)
      pre_period = [0, 69]
      post_period = [70, 99]

      ci = CausalImpact(data, pre_period, post_period)
      print(ci.summary())
      print(ci.summary('report'))
      ci.plot()
      ```

      Using pandas DataFrames:

      ```python
      df = pd.DataFrame('tests/fixtures/arma_data.csv')
      pre_period = [0, 69]
      post_period = [70, 99]
      ci = CausalImpact(df, pre_period, post_period)
      print(ci.summary())
      ```

      Using pandas DataFrames with pandas timestamps:

      ```python
      df = pd.DataFrame('tests/fixtures/arma_data.csv')
      df = df.set_index(pd.date_range(start='20200101', periods=len(data)))
      pre_period = [pd.to_datetime('20200101'), pd.to_datetime('20200311')]
      post_period = [pd.to_datetime('20200312'), pd.to_datetime('20200410')]
      ci = CausalImpact(df, pre_period, post_period)
      print(ci.summary())
      ```

      Using a weekly seasonal component on daily data:

      ```python
      df = pd.DataFrame('tests/fixtures/arma_data.csv')
      df = df.set_index(pd.date_range(start='20200101', periods=len(data)))
      pre_period = ['20200101', '20200311']
      post_period = ['20200312', '20200410']
      ci = CausalImpact(df, pre_period, post_period, model_args={'nseasons': 7})
      print(ci.summary())
      ```

      Using a weekly seasonal component on hourly data:

      ```python
      df = pd.DataFrame('tests/fixtures/arma_data.csv')
      df = df.set_index(pd.date_range(start='20200101', periods=len(data), freq='H'))
      pre_period = ['20200101 00:00:00', '20200311 23:00:00']
      post_period = ['20200312 00:00:00', '20200410 23:00:00']
      ci = CausalImpact(df, pre_period, post_period, model_args={'nseasons': 7,
                        'season_duration': 24})
      print(ci.summary())
      ```

      Using a customized model:

      ```python
      import tensorflow_probability as tfp


      pre_y = data[:70, 0]
      pre_X = data[:70, 1:]
      obs_series = data.iloc[:, 0]
      local_linear = tfp.sts.LocalLinearTrend(observed_time_series=obs_series)
      seasonal = tfp.sts.Seasonal(nseasons=7, observed_time_series=obs_series)
      model = tfp.sts.Sum([local_linear, seasonal], observed_time_series=obs_series)

      ci = CausalImpact(data, pre_period, post_period, model=model)
      print(ci.summary())
      ```
    """
    def __init__(
        self,
        data: Union[np.array, pd.DataFrame],
        pre_period: Union[List[int], List[str], List[pd.Timestamp]],
        post_period: Union[List[int], List[str], List[pd.Timestamp]],
        model: Optional[tfp.sts.StructuralTimeSeries] = None,
        model_args: Dict[str, Any] = {},
        alpha: float = 0.05
    ):
        processed_input = cidata.process_input_data(data, pre_period, post_period,
                                                    model, model_args, alpha)
        self.data = data
        self.pre_period = processed_input['pre_period']
        self.post_period = processed_input['post_period']
        self.pre_data = processed_input['pre_data']
        self.post_data = processed_input['post_data']
        self.alpha = processed_input['alpha']
        self.model_args = processed_input['model_args']
        self.model = processed_input['model']
        self.normed_pre_data = processed_input['normed_pre_data']
        self.normed_post_data = processed_input['normed_post_data']
        self.observed_time_series = processed_input['observed_time_series']
        self.mu_sig = processed_input['mu_sig']
        self._fit_model()
        self._process_posterior_inferences()
        self._summarize_inferences()

    def plot(
        self,
        panels: List[str] = ['original', 'pointwise', 'cumulative'],
        figsize: Tuple[int] = (10, 7),
        show: bool = True
    ) -> None:
        """
        Plots the graphic with results associated to Causal Impact.

        Args
        ----
          panels: List[str]
              Which graphics to plot. 'original' plots the original data, forecasts means
              and credible intervals related to the fitted model.
              'pointwise' plots the point wise differences between observed data and
              predictions. Finally, 'cumulative' is a cumulative summation over real
              data and its forecasts.
          figsize: Tuple[int]
              Sets the width and height of the figure to plot.
          show: bool
            If `True` then plots the figure by running `plt.plot()`.
            If `False` then nothing will be plotted which allows for accessing and
            manipulating the figure and axis of the plot, i.e., the figure can be saved
            and the styling can be modified. To get the axis, just run:
            `import matplotlib.pyplot as plt; ax = plt.gca()` or the figure:
            `fig = plt.gcf()`. Defaults to `True`.
        """
        plotter.plot(self.inferences, self.pre_data, self.post_data, panels=panels,
                     figsize=figsize, show=show)

    def summary(self, output: str = 'summary', digits: int = 2) -> str:
        """
        Builds and prints the summary report.

        Args
        ----
          output: str
              Can be either "summary" or "report". The first is a simpler output just
              informing general metrics such as expected absolute or relative effect.
          digits: int
              Defines the number of digits after the decimal point to round. For
              `digits=2`, value 1.566 becomes 1.57.

        Returns
        -------
          summary: str
              Contains results of the causal impact analysis.

        Raises
        ------
          ValueError: If input `output` is not either 'summary' or 'report'.
                      If input `digits` is not of type integer.
        """
        if not isinstance(digits, int):
            raise ValueError(
                f'Input value for digits must be integer. Received "{type(digits)}" '
                'instead.'
            )
        result = summarizer.summary(self.summary_data, self.p_value, self.alpha,
                                    output, digits)
        return result

    def _fit_model(self) -> None:
        """
        Use observed data `Y` to find the posterior `P(Z|Y)` where `Z` represents the
        structural components that were used for building the model (such as local level
        factor or seasonal components).
        """
        model_samples, model_kernel_results = cimodel.fit_model(
            self.model,
            self.observed_time_series,
            self.model_args['fit_method'],
        )
        self.model_samples = model_samples
        self.model_kernel_results = model_kernel_results

    def _summarize_inferences(self) -> None:
        """
        After processing predictions and forecasts, use these values to build the
        summary data used for reporting and plotting. Computes the estimated p-value
        for determining if the impact is statistically significant or not.
        """
        post_preds_means = self.inferences['post_preds_means']
        post_data_sum = self.post_data.iloc[:, 0].sum()
        niter = self.model_args['niter']
        simulated_ys = maybe_unstandardize(
            np.squeeze(self.posterior_dist.sample(niter).numpy()),
            self.mu_sig
        )
        self.summary_data = inferrer.summarize_posterior_inferences(post_preds_means,
                                                                    self.post_data,
                                                                    simulated_ys,
                                                                    self.alpha)
        self.p_value = inferrer.compute_p_value(simulated_ys, post_data_sum)

    def _process_posterior_inferences(self) -> None:
        """
        Run `inferrer` to process data forecasts and predictions. Results feeds the
        summary table as well as the plotting functionalities.
        """
        num_steps_forecast = len(self.post_data)
        self.one_step_dist = cimodel.build_one_step_dist(self.model,
                                                         self.observed_time_series,
                                                         self.model_samples)
        self.posterior_dist = cimodel.build_posterior_dist(self.model,
                                                           self.observed_time_series,
                                                           self.model_samples,
                                                           num_steps_forecast)
        self.inferences = inferrer.compile_posterior_inferences(
            self.pre_data,
            self.post_data,
            self.one_step_dist,
            self.posterior_dist,
            self.mu_sig,
            self.alpha,
            self.model_args['niter']
        )



# Copyright WillianFuks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module responsible for processing data that feeds the causal impact algorithm.
"""


from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import tensorflow_probability as tfp

import causalimpact.model as cimodel
from causalimpact.misc import standardize


def process_input_data(
    data: Union[np.array, pd.DataFrame],
    pre_period: Union[List[int], List[str], List[pd.Timestamp]],
    post_period: Union[List[int], List[str], List[pd.Timestamp]],
    model: Optional[tfp.sts.StructuralTimeSeries],
    model_args: Dict[str, Any],
    alpha: float
) -> Dict[str, Any]:
    """
    Checks and formats input data for running the Causal Impact algorithm.

    Args
    ----
      data: Union[np.array, pd.DataFrame]
          First column is the response variable `y` and other columns correspond to
          the covariates `X`.
      pre_period: Union[List[int], List[str], List[pd.Timestamp]]
          List with initial and final points to consider for pre-intervention period.
      post_period: Union[List[int], List[str], List[pd.Timestamp]]
          List with initial and final points to consider for post-intervention period.
      model: Optional[tfd.LinearGaussianStateSpaceModel]
          If `None` then uses default `tfp.sts.LocalLevel` model otherwise it should
          be a `tfd.LinearGaussianStateSpaceModel` specified by client.
      model_args: Dict[str, Any]
          Sets general variables for building and running the state space model. Possible
          values are:
            niter: int
                Number of samples to draw from posterior `P(z | y)`.
            standardize: bool
                If `True`, standardizes data to have zero mean and unitary standard
                deviation.
            prior_level_sd: Optional[float]
                Prior value for the local level standard deviation. If `None` then an
                automatic optimization of the local level is performed. This is
                recommended when there's uncertainty about what prior value is
                appropriate for the data.
                In general, if the covariates are expected to be good descriptors of the
                observed response then this value can be low (such as the default of
                0.01). In cases when the linear regression is not quite expected to fully
                explain the observed data, the value 0.1 can be used.
            fit_method: str
                Which method to use for the Bayesian algorithm. Can be either 'vi'
                (default) or 'hmc' (more precision but much slower).
            nseasons: int
              Specifies the duration of the period of the seasonal component; if input
              data is specified in terms of days, then choosing nseasons=7 adds a weekly
              seasonal effect.
            season_duration: int
              Specifies how many data points each value in season spans over. A good
              example to understand this argument is to consider a hourly data as input.
              For modeling a weekly season on this data, one can specify `nseasons=7` and
              season_duration=24 which means each value that builds the season component
              is repeated for 24 data points. Default value is 1 which means the season
              component spans over just 1 point (this in practice doesn't change
              anything). If this value is specified and bigger than 1 then `nseasons`
              must be specified and bigger than 1 as well.
      alpha: float
          Used for two-sided statistical test on comparison between counter-factual
          and observed series.

    Returns
    -------
      Dict[str, Any]:
        data: pd.DataFrame
            Validated data, first column is `y` and the others are the `X` covariates.
        pre_period: List[int]
        post_period: List[int]
        pre_data: pd.DataFrame
            Data sliced using `pre_period` values.
        post_data: pd.DataFrame
        normed_pre_data: pd.DataFrame
            If `standardize==True` then this is the result of the input data being
            normalized.
        observed_time_series: pd.DataFrame
            The input time series used for the TFP API. Main change is the index now
            is mandatory to be of type `DatetimeIndex` and have a valid frequency.
        normed_post_data: pd.DataFrame
        model: tfp.sts.StructuralTimeSeries
            `tfp.sts.StructuralTimeSeries` validated input model.
        model_args: Dict[str, Any]
            Dict containing general information related to how to fit and run the
            structural time series model.
        alpha: float
        mu_sig: Tuple[float, float]
            Mean and standard deviation used to normalize just the response variable `y`.

    Raises
    ------
      ValueError: if input arguments is `None`.
    """
    _check_empty_inputs(locals())
    fmt_data = format_input_data(data)
    pre_data, post_data = process_pre_post_data(fmt_data, pre_period, post_period)
    alpha = process_alpha(alpha)
    model_args = cimodel.process_model_args(model_args if model_args else {})
    normed_data = (standardize_pre_and_post_data(pre_data, post_data) if
                   model_args['standardize'] else (None, None, None))
    normed_pre_data, normed_post_data = normed_data[0], normed_data[1]
    observed_time_series = build_observed_time_series(
        pre_data if normed_pre_data is None else normed_pre_data
    )
    if model:
        cimodel.check_input_model(model, pre_data, post_data)
    else:
        model = cimodel.build_default_model(
            observed_time_series,
            normed_pre_data if model_args['standardize'] else pre_data,
            normed_post_data if model_args['standardize'] else post_data,
            model_args['prior_level_sd'],
            model_args['nseasons'],
            model_args['season_duration']
        )
    return {
        'data': fmt_data,
        'pre_period': pre_period,
        'post_period': post_period,
        'pre_data': pre_data,
        'post_data': post_data,
        'normed_pre_data': normed_data[0],
        'normed_post_data': normed_data[1],
        'observed_time_series': observed_time_series,
        'model': model,
        'model_args':  model_args,
        'alpha': alpha,
        'mu_sig': normed_data[2]
    }


def build_observed_time_series(pre_data: pd.DataFrame) -> pd.DataFrame:
    """
    Helper method for creating the observed time series data to be used as input for
    the TFP API. If input data index is of type
    `pandas.core.indexes.range.RangeIndex`
    then it's casted to `DatetimeIndex` as required by `TFP==0.14.0`. A dummy date index
    is created starting from '2020-01-01'.
    """
    # type must be cast to `np.float32` as the linear regressor from tensorflow only
    # works with 32 bytes.
    observed_time_series = pre_data.astype(np.float32)
    # if operation `iloc` returns a pd.Series, cast it back to pd.DataFrame
    observed_time_series = pd.DataFrame(observed_time_series.iloc[:, 0])
    if isinstance(pre_data.index, pd.RangeIndex):
        observed_time_series.set_index(
            pd.date_range(start='2020-01-01', periods=len(observed_time_series)),
            inplace=True
        )
    return tfp.sts.regularize_series(observed_time_series)


def _check_empty_inputs(inputs: Dict[str, Any]) -> None:
    """
    Works as a validator for whether input values for Causal Impact contains empty
    values.

    Args
    ----
      inputs: Dict[str, Any]
          Basically contains the `locals()` variables as received by CausalImpact.
    """
    none_args = sorted(
        [arg for arg in inputs if inputs[arg] is None and arg != 'model']
    )
    if none_args:
        raise ValueError(
            f'{", ".join(none_args)} '
            f'input argument{"s" if len(none_args) > 1 else ""} cannot be empty'
        )


def standardize_pre_and_post_data(
    pre_data: pd.DataFrame,
    post_data: pd.DataFrame
) -> Tuple[pd.DataFrame, pd.DataFrame, Tuple[float, float]]:
    """
    Applies standardization in pre and post data, based on mean and standard deviation
    of `pre_data` (as it's used for training the causal impact model).

    Args
    ----
      pre_data: pd.DataFrame
          data selected to be the pre-intervention dataset of causal impact.
      post_data: pd.DataFrame

    Returns
    -------
      Tuple[pd.DataFrame, pd.DataFrame, Tuple[float, float]]
        `pre_data` and `post_data` normalized along with the mean and variance used for
        response variable `y` only.
    """
    normed_pre_data, (mu, sig) = standardize(pre_data)
    normed_post_data = (post_data - mu) / sig
    mu_sig = (mu.iloc[0], sig.iloc[0])
    return (normed_pre_data, normed_post_data, mu_sig)


def format_input_data(data: Union[np.array, pd.DataFrame]) -> pd.DataFrame:
    """
    Validates and formats input data.

    Args
    ----
      data: Union[np.array, pd.DataFrame]
          First column is the response variable `y` and other columns correspond to
          the covariates `X`.

    Returns
    -------
      data: pd.DataFrame
          Validated data to be used in Causal Impact algorithm.

    Raises
    ------
      ValueError: if input `data` is non-convertible to pandas DataFrame.
                  if input `data` has non-numeric values.
                  if input `data` has less than 3 points.
                  if input covariates have NAN values.
    """
    if not isinstance(data, pd.DataFrame):
        try:
            data = pd.DataFrame(data)
        except ValueError:
            raise ValueError(
                'Could not transform input data to pandas DataFrame.'
            )
    validate_y(data.iloc[:, 0])
    # must contain only numeric values
    if not data.applymap(np.isreal).values.all():
        raise ValueError('Input data must contain only numeric values.')
    # covariates cannot have NAN values
    if data.shape[1] > 1:
        if data.iloc[:, 1:].isna().values.any():
            raise ValueError('Input data cannot have NAN values.')
    # If index is a string of dates, try to convert it to datetimes which helps
    # in plotting.
    data = convert_index_to_datetime(data)
    return data


def process_pre_post_data(
    data: pd.DataFrame,
    pre_period: Union[List[int], List[str], List[pd.Timestamp]],
    post_period: Union[List[int], List[str], List[pd.Timestamp]]
) -> List[pd.DataFrame]:
    """
    Checks `pre_period` and `post_period` to return input data sliced accordingly to
    each period.

    Args
    ----
      data: pd.DataFrame
          First column is the response variable `y` and other columns correspond to
          the covariates `X`. This input data has already been validated as appropriate
          for the causal impact algorithm.
      pre_period: Union[List[int], List[str], List[pd.Timestamp]]
      post_period: Union[List[int], List[str], List[pd.Timestamp]]

    Returns
    -------
      result: List[pd.DataFrame]
          First value is pre-intervention data and second value is post-intervention.

    Raises
    ------
      ValueError: if pre_period last value is bigger than post intervention period.
    """
    checked_pre_period = process_period(pre_period, data)
    checked_post_period = process_period(post_period, data)
    if checked_pre_period[1] > checked_post_period[0]:
        raise ValueError(
            'Values in training data cannot be present in the post-intervention '
            'data. Please fix your pre_period value to cover at most one point less '
            'from when the intervention happened.'
        )
    if checked_pre_period[1] < checked_pre_period[0]:
        raise ValueError('pre_period last number must be bigger than its first.')
    if checked_pre_period[1] - checked_pre_period[0] < 3:
        raise ValueError('pre_period must span at least 3 time points.')
    if checked_post_period[1] < checked_post_period[0]:
        raise ValueError('post_period last number must be bigger than its first.')
    if checked_post_period[0] <= checked_pre_period[1]:
        raise ValueError(f'post_period first value ({post_period[0]}) must '
                         'be bigger than the second value of pre_period '
                         f'({pre_period[1]}).')
    result = [
        data.loc[pre_period[0]: pre_period[1], :],
        data.loc[post_period[0]: post_period[1], :]
    ]
    return result


def validate_y(y: pd.Series) -> None:
    """
    Validates if input response variable is correct and doesn't contain invalid input.

    Args
    ----
      y: pd.Series
          Response variable sent in input data in first column.

    Raises
    ------
      ValueError: if values in `y` are Null.
                  if less than 3 (three) non-null values in `y` (as in this case
                      we can't even train a model).
                  if `y` is constant (in this case it doesn't make much sense to
                    make predictions as the time series doesn't change in the
                    training phase.
    """
    if np.all(y.isna()):
        raise ValueError('Input response cannot have just Null values.')
    if y.notna().values.sum() < 3:
        raise ValueError('Input response must have more than 3 non-null '
                         'points at least.')
    if y.std(skipna=True, ddof=0) == 0:
        raise ValueError('Input response cannot be constant.')


def convert_index_to_datetime(data: pd.DataFrame) -> pd.DataFrame:
    """
    If input data has index of string dates, i.e, '20200101', '20200102' and so on, try
    to convert it to datetime specifically, which results in
    Timestamp('2020-01-01 00:00:00'), Timestamp('2020-01-02 00:00:00') ...

    Args
    ----
      data: pd.DataFrame
          Input data used in causal impact analysis.

    Returns
    -------
      data: pd.DataFrame
          Same input data with potentially new index of type DateTime.
    """
    if isinstance(data.index.values[0], str):
        try:
            data.set_index(pd.to_datetime(data.index), inplace=True)
        except ValueError:
            pass
    return data


def process_period(
    period: Union[List[int], List[str], List[pd.Timestamp]],
    data: pd.DataFrame
) -> List[int]:
    """
    Validates period inputs.

    Args
    ----
      period: Union[List[int], List[str], List[pd.Timestamp]]
      data: pd.DataFrame.
          Input Causal Impact data.

    Returns
    -------
      period: List[int]
          Validated period list.

    Raises
    ------
      ValueError: if input `period` is not of type list.
                  if input doesn't have two elements.
                  if period date values are not present in data.
    """
    if not isinstance(period, list):
        raise ValueError('Input period must be of type list.')
    if len(period) != 2:
        raise ValueError(
            'Period must have two values regarding the beginning and end of '
            'the pre and post intervention data.'
        )
    none_args = [d for d in period if d is None]
    if none_args:
        raise ValueError('Input period cannot have `None` values.')
    if not (
        (isinstance(period[0], int) and isinstance(period[1], int)) or
        (isinstance(period[1], str) and isinstance(period[1], str)) or
        (isinstance(period[1], pd.Timestamp) and isinstance(period[1], pd.Timestamp))
    ):
        raise ValueError(
            'Input periods must contain either int, str or pandas Timestamp'
        )
    # check whether the input period is indeed present in the input data index
    for point in period:
        if point not in data.index:
            if isinstance(point, pd.Timestamp):
                point = point.strftime('%Y%m%d')
            raise ValueError("{point} not present in input data index.".format(
                point=str(point)
                )
            )
    if isinstance(period[0], str) or isinstance(period[0], pd.Timestamp):
        period = convert_date_period_to_int(period, data)
    return period


def convert_date_period_to_int(
    period: Union[List[int], List[str], List[pd.Timestamp]],
    data: pd.DataFrame
) -> List[int]:
    """
    Converts string values from `period` to integer offsets from `data`.

    Args
    ----
      period: Union[List[int], List[str], List[pd.Timestamp]]
      data: pd.DataFrame
          `data` index is either a `str` of a pd.Timestamp type.

    Returns
    -------
      period: List[int]
          Where each value is the correspondent integer based value in `data` index.
    """
    result = []
    for date in period:
        offset = data.index.get_loc(date)
        result.append(offset)
    return result


def process_alpha(alpha: float) -> float:
    """
    Asserts input `alpha` is appropriate to be used in the model.

    Args
    ----
      alpha: float
          Ranges from 0 up to 1 indicating level of significance to assert when
          testing for presence of signal in post-intervention data.

    Returns
    -------
      alpha: float
          Validated `alpha` value.

    Raises
    ------
      ValueError: if alpha is not float.
                  if alpha is not between 0. and 1.
    """
    if not isinstance(alpha, float):
        raise ValueError('alpha must be of type float.')
    if alpha < 0 or alpha > 1:
        raise ValueError(
            'alpha must range between 0 (zero) and 1 (one) inclusive.'
        )
    return alpha



# Copyright WillianFuks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import sys

from setuptools import setup


here = os.path.abspath(os.path.dirname(__file__))

if sys.argv[-1] == 'test-publish':
    """
    Publishes to test pypi repository.
    """
    if os.path.exists('dist') or os.path.exists('build'):
        raise RuntimeError('Please first delete dist/build folders')
    os.system('pip install -U twine')
    os.system('sh ./scripts/build_wheels.sh')
    os.system('twine upload --repository testpypi dist/*')
    # os.system('rm -rf build dist .egg *.egg-info')
    sys.exit()

if sys.argv[-1] == 'publish':
    """
    Publishes to original pypi repository.
    """
    if os.path.exists('dist') or os.path.exists('build'):
        raise RuntimeError('Please first delete dist/build folders')
    os.system('pip install -U twine')
    os.system('sh ./scripts/build_wheels.sh')
    os.system('twine upload dist/*')
    # os.system('rm -rf build dist .egg *.egg-info')
    sys.exit()

install_requires = [
    'jinja2',
    'pandas',
    'tensorflow',
    'tensorflow-probability',
    'matplotlib',
]
tests_require = [
    'pytest',
    'pytest-cov',
    'mock',
    'tox'
]
setup_requires = [
    'flake8',
    'isort'
]
extras_require = {
    'docs': [
        'ipython',
        'jupyter'
    ]
}

packages = ['causalimpact']

_version = {}
_version_path = os.path.join(here, 'causalimpact', '__version__.py')

with open(_version_path, 'r') as f:
    exec(f.read(), _version)

with open('README.md', 'r') as f:
    readme = f.read()

setup(
    name='tfcausalimpact',
    version=_version['__version__'],
    author='Willian Fuks',
    author_email='willian.fuks@gmail.com',
    url='https://github.com/WillianFuks/tfcausalimpact',
    description= "Python version of Google's Causal Impact model on top of Tensorflow Probability.",
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=packages,
    include_package_data=True,
    install_requires=install_requires,
    tests_require=tests_require,
    setup_requires=setup_requires,
    extras_require=extras_require,
    license='Apache License 2.0',
    keywords='causal impact',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Scientific/Engineering',
    ],
    project_urls={
        'Source': 'https://github.com/WillianFuks/tfcausalimpact'
    },
    python_requires='>=3, <3.10.*',
    test_suite='tests'
)



# Copyright WillianFuks
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Plots the analysis obtained in causal impact algorithm.
"""


import pandas as pd


def plot(
    inferences: pd.DataFrame,
    pre_data: pd.DataFrame,
    post_data: pd.DataFrame,
    panels=['original', 'pointwise', 'cumulative'],
    figsize=(10, 7),
    show=True
) -> None:
    """Plots inferences results related to causal impact analysis.

    Args
    ----
      panels: list.
        Indicates which plot should be considered in the graphics.
      figsize: tuple.
        Changes the size of the graphics plotted.
      show: bool.
        If true, runs plt.show(), i.e., displays the figure.
        If false, it gives acess to the axis, i.e., the figure can be saved
        and the style of the plot can be modified by getting the axis with
        `ax = plt.gca()` or the figure with `fig = plt.gcf()`.
        Defaults to True.
    Raises
    ------
      RuntimeError: if inferences were not computed yet.
    """
    plt = get_plotter()
    plt.figure(figsize=figsize)
    valid_panels = ['original', 'pointwise', 'cumulative']
    for panel in panels:
        if panel not in valid_panels:
            raise ValueError(
                '"{}" is not a valid panel. Valid panels are: {}.'.format(
                    panel, ', '.join(['"{}"'.format(e) for e in valid_panels])
                )
            )
    pre_post_index = pre_data.index.union(post_data.index)
    post_period_init = post_data.index[0]
    intervention_idx = pre_post_index.get_loc(post_period_init)
    n_panels = len(panels)
    ax = plt.subplot(n_panels, 1, 1)
    idx = 1
    color = (1.0, 0.4981, 0.0549)
    # The operation `iloc[1:]` is used mainly to remove the uncertainty associated to the
    # predictions of the first points. As the predictions follow
    # `P(z[t] | y[1...t-1], z[1...t-1])` the very first point ends up being quite noisy
    # as there's no previous point observed.
    if 'original' in panels:
        ax.plot(
            pre_post_index,
            pd.concat([pre_data.iloc[:, 0], post_data.iloc[:, 0]]),
            'k',
            label='y'
        )
        ax.plot(
            pre_post_index[1:],
            inferences['complete_preds_means'].iloc[1:],
            color='orangered',
            ls='dashed',
            label='Predicted'
        )
        ax.axvline(pre_post_index[intervention_idx - 1], c='gray', linestyle='--')
        ax.fill_between(
            pre_post_index[1:],
            inferences['complete_preds_lower'].iloc[1:],
            inferences['complete_preds_upper'].iloc[1:],
            color=color,
            alpha=0.4
        )
        ax.legend()
        ax.grid(True, color='gainsboro')
        if idx != n_panels:
            plt.setp(ax.get_xticklabels(), visible=False)
        idx += 1
    if 'pointwise' in panels:
        ax = plt.subplot(n_panels, 1, idx, sharex=ax)
        ax.plot(
            pre_post_index[1:],
            inferences['point_effects_means'].iloc[1:],
            ls='dashed',
            color='orangered',
            label='Point Effects'
        )
        ax.axvline(pre_post_index[intervention_idx - 1], c='gray', linestyle='--')
        ax.fill_between(
            pre_post_index[1:],
            inferences['point_effects_lower'].iloc[1:],
            inferences['point_effects_upper'].iloc[1:],
            color=color,
            alpha=0.4
        )
        ax.axhline(y=0, color='gray')
        ax.legend()
        ax.grid(True, color='gainsboro')
        if idx != n_panels:
            plt.setp(ax.get_xticklabels(), visible=False)
        idx += 1
    if 'cumulative' in panels:
        ax = plt.subplot(n_panels, 1, idx, sharex=ax)
        ax.plot(
            pre_post_index[1:],
            inferences['post_cum_effects_means'].iloc[1:],
            ls='dashed',
            color='orangered',
            label='Cumulative Effect'
        )
        ax.axvline(pre_post_index[intervention_idx - 1], c='gray', linestyle='--')
        ax.fill_between(
            pre_post_index[1:],
            inferences['post_cum_effects_lower'].iloc[1:],
            inferences['post_cum_effects_upper'].iloc[1:],
            color=color,
            alpha=0.4
        )
        ax.axhline(y=0, color='gray', linestyle='--')
        ax.legend()
        ax.grid(True, color='gainsboro')
    if show:
        plt.show()


def get_plotter():  # pragma: no cover
    """As some environments do not have matplotlib then we import the library through
    this method which prevents import exceptions.

    Returns
    -------
      plotter: `matplotlib.pyplot`.
    """
    import matplotlib.pyplot as plt
    return plt
